/*
 This file is part of GNU Taler
 (C) 2021-2023 Taler Systems S.A.

 GNU Taler is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
 Foundation; either version 3, or (at your option) any later version.

 GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */

/**
 *
 * @author Sebastian Javier Marchano (sebasjm)
 */

import {
  ErrorType,
  HttpError,
  useTranslationContext,
} from "@gnu-taler/web-util/browser";
import { Fragment, h, VNode } from "preact";
import { useState } from "preact/hooks";
import { Loading } from "../../../../components/exception/loading.js";
import { NotificationCard } from "../../../../components/menu/index.js";
import { MerchantBackend, WithId } from "../../../../declaration.js";
import { Notification } from "../../../../utils/types.js";
import { UpdatePage } from "./UpdatePage.js";
import { HttpStatusCode } from "@gnu-taler/taler-util";
import { useOtpDeviceAPI, useOtpDeviceDetails } from "../../../../hooks/otp.js";
import { CreatedSuccessfully } from "../create/CreatedSuccessfully.js";

export type Entity = MerchantBackend.OTP.OtpDevicePatchDetails & WithId;

interface Props {
  onBack?: () => void;
  onConfirm: () => void;
  onUnauthorized: () => VNode;
  onNotFound: () => VNode;
  onLoadError: (e: HttpError<MerchantBackend.ErrorDetail>) => VNode;
  vid: string;
}
export default function UpdateValidator({
  vid,
  onConfirm,
  onBack,
  onUnauthorized,
  onNotFound,
  onLoadError,
}: Props): VNode {
  const { updateOtpDevice } = useOtpDeviceAPI();
  const result = useOtpDeviceDetails(vid);
  const [notif, setNotif] = useState<Notification | undefined>(undefined);
  const [keyUpdated, setKeyUpdated] = useState<MerchantBackend.OTP.OtpDeviceAddDetails | null>(null)

  const { i18n } = useTranslationContext();

  if (result.loading) return <Loading />;
  if (!result.ok) {
    if (
      result.type === ErrorType.CLIENT &&
      result.status === HttpStatusCode.Unauthorized
    )
      return onUnauthorized();
    if (
      result.type === ErrorType.CLIENT &&
      result.status === HttpStatusCode.NotFound
    )
      return onNotFound();
    return onLoadError(result);
  }

  if (keyUpdated) {
    return <CreatedSuccessfully entity={keyUpdated} onConfirm={onConfirm} />
  }

  return (
    <Fragment>
      <NotificationCard notification={notif} />
      <UpdatePage
        device={{
          id: vid,
          otp_algorithm: result.data.otp_algorithm,
          otp_device_description: result.data.device_description,
          otp_key: undefined,
          otp_ctr: result.data.otp_ctr
        }}
        onBack={onBack}
        onUpdate={async (newInfo) => {
          return updateOtpDevice(vid, newInfo)
            .then((d) => {
              if (newInfo.otp_key) {
                setKeyUpdated({
                  otp_algorithm: newInfo.otp_algorithm,
                  otp_device_description: newInfo.otp_device_description,
                  otp_device_id: newInfo.id,
                  otp_key: newInfo.otp_key,
                  otp_ctr: newInfo.otp_ctr,
                })
              } else {
                onConfirm()
              }
            })
            .catch((error) => {
              setNotif({
                message: i18n.str`could not update template`,
                type: "ERROR",
                description: error.message,
              });
            });
        }}
      />
    </Fragment>
  );
}
